;;; documented versions of some interfaces in interfaces.scm

(define-documented-interface swank-worlds-interface
  "A Swank world is a contained environments for interaction sessions.
   It contains a general scratch environment for user interaction, a
   config environment for the module language and all structure &
   interface definitions, and a Swank RPC environment to evaluate Swank
   RPC expressions from the remote system.

   This interface exports operations on Swank worlds, facilities to
   list & extract packages & structures from the config environment,
   and a utility for constructing the environments for a Swank world."
  (export ((MAKE-SWANK-WORLD scratch-env config-env rpc-env id)
           :swank-world
           "Swank world constructor.  ID is used only for debugging
            purposes.  SCRATCH-ENV, CONFIG-ENV, and RPC-ENV are
            typically obtained by a call to `MAKE-SWANK-ENVS'.")
          ((SWANK-WORLD? obj) :boolean
           "Disjoint type predicate for Swank worlds.")
          ((SWANK-WORLD-ID world) :value
           "Returns WORLD's identifier as passed to `MAKE-SWANK-WORLD'
            to construct it.")
          ((SWANK-WORLD-SESSIONS world) :weak-set
           "Returns a weak set of all sessions running in WORLD.")
          ((SWANK-WORLD-SCRATCH-ENV world) :package
           "Returns WORLD's scratch environment for user interaction.")
          ((SWANK-WORLD-CONFIG-ENV world) :package
           "Returns WORLD's config environment, suitable for evaluating
            structure and interface definitions.")
          ((SWANK-WORLD-RPC-ENV world) :package
           "Returns WORLD's environment for Swank RPC evaluations.")
          ((SWANK-WORLD-STRUCTURE-NAMES world) :list
           "Returns a list of all names defined as structures in the
            config environment of WORLD.  The names are suitable for
            passing to `FIND-STRUCTURE-IN-SWANK-WORLD'.")
          ((SWANK-WORLD-PACKAGE-NAMES world) :list
           "Returns a list of all names defined in WORLD's config
            environment as structures with unstable packages, i.e.
            packages into which new code may be loaded.  The names are
            suitable for passing to `FIND-PACKAGE-IN-SWANK-WORLD'.")
          ((FIND-STRUCTURE-IN-SWANK-WORLD structure-specifier world)
           :structure
           "Finds a structure defined in WORLD's config environment for
            STRUCTURE-SPECIFIER, which should be a symbol.  If no such
            structure exists, returns `#F'.")
          ((FIND-PACKAGE-IN-SWANK-WORLD specifier world)
           :package
           "Finds a package named by SPECIFIER in WORLD.  SPECIFIER may
            take four possible forms: if it is the list `(SCRATCH)', it
            denotes WORLD's scratch environment; likewise with the list
            `(CONFIG)' & WORLD's config environment, and `(RPC)' &
            WORLD's RPC environment; or if it is a symbol, and it is
            bound to a structure in WORLD's config environment, that
            structure's package is returned.  If it satisfies none of
            these cases, `FIND-PACKAGE-IN-SWANK-WORLD' returns `#F'.")
          ((MAKE-SWANK-ENVS scheme-struct
                            config-lang-struct
                            built-in-struct
                            rpcs-struct)
           (:values :package :package :package)
           "Creates the three environments for a Swank world.  The
            scratch and config environments are created with a
            reflective tower in all of whose storeys the environment is
            a simple package that opens SCHEME-STRUCT and the evaluator
            is `(SCHEME)EVAL'.  The scratch environment is a simple,
            unstable package that opens SCHEME-STRUCT; it is also set
            never to integrate, so primitives may be redefined at whim.
            The config environment is created as if by
            `(MODULE-CONTROL)MAKE-CONFIG-PACKAGE'.  The RPC environment
            is created to import only `QUOTE' from SCHEME-STRUCT, and
            all names from RPCS-STRUCT.")))

(define-documented-interface swank-logging-interface
  "Simple formatted logging facility for use by Swank."
  (export ((SWANK-LOG format . args)
           "Outputs a log entry with the given format string and
            arguments, formatted by `(FORMATS)FORMAT'.")))

(define-documented-interface package-loader-interface
  "Package loader."
  (export ((LOAD-PACKAGE package)
           "Loads PACKAGE's code, compiling it if necessary, after
            first recursively ensuring that the packages of all
            structures opened or accessed by PACKAGE are loaded.")))

(define-documented-interface restarting-interface
  "A simple Common-Lisp-style condition recovery system.  /Restarters/
   are objects that represent methods by which to restart an
   exceptional situation, possibly interactively, with a human-readable
   message describing the method and a symbolic tag.  Restarters may be
   non-interactive, in which case they have `#F' specified instead of
   an interactor procedure."
  (export ((MAKE-RESTARTER tag description invoker interactor)
           :restarter
           "Restarter constructor.")
          ((RESTARTER? obj) :boolean
           "Disjoint type predicate for restarters.")
          ((RESTARTER-TAG restarter) :value
           "Returns RESTARTER's tag.")
          ((RESTARTER-DESCRIPTION restarter) :string
           "Returns RESTARTER's human-readable description string.")
          ((RESTART restarter . arguments)
           "Invokes RESTARTER with the given arguments.  This may not
            return.")
          ((RESTART-INTERACTIVELY restarter)
           "Invokes RESTARTER interactively: runs its interactor, and
            then passes the values that the interactor returns to the
            invoker procedure as if with `RESTART'.")
          ((CURRENT-RESTARTERS) :list
           "Returns the list of current restarters.  It is an error to
            modify this list.")
          ((RESTARTERS-IN-THREAD thread) :list
           "Returns a list of the current restarters in THREAD's
            dynamic environment.  This is intended mainly for debuggers
            to use.")
          ((WITH-RESTARTER restarter thunk) :values
           ;; (put 'with-restarter 'scheme-indent-function 1)
           "Applies THUNK to zero arguments, with RESTARTER pushed onto
            the list of current restarters for the dynamic extent of
            THUNK.")
          ((FIND-RESTARTER tag) (false-or :restarter)
           "Returns the first restarter in the current list of
            whose tag is TAG, or `#F' if no restarter has such a tag.
            Restarter tags are compared with `(SCHEME)EQV?'.")
          ((CALL-WITH-RESTARTER tag description invoker receiver)
           :values
           ;; (put 'call-with-restarter 'scheme-indent-function 3)
           "Constructs a non-interactive restarter with the given tag,
            description, and invoker procedure, and applies RECEIVER to
            one argument, that restarter, with that restarter pushed
            onto the current list of restarters for the dynamic extent
            of RECEIVER.")
          ((CALL-WITH-INTERACTIVE-RESTARTER tag description
               invoker interactor
             receiver)
           :values
           ;; (put 'call-with-interactive-restarter 'scheme-indent-function 4)
           "Constructs an interactive restarter with the given fields
            and applies RECEIVER to one argument, that restarter, with
            that restarter pushed onto the current list of restarters
            for the dynamic extent of RECEIVER.")
          ((WITH-EXITING-RESTARTER tag description thunk) :values
           ;; (put 'with-exiting-restarter 'scheme-indent-function 2)
           "Creates an interactive restarter whose invoker accepts zero
            arguments and returns zero values to the original call to
            `WITH-EXITING-RESTARTER' and whose interactor returns zero
            values; and calls THUNK with zero arguments, for whose
            dynamic extent that restarter is pushed onto the current
            list of restarters.")
          ((CALL-WITH-EXITING-RESTARTER tag description receiver)
           :values
           ;; (put 'call-with-exiting-restarter 'scheme-indent-function 2)
           "Like `WITH-EXITING-RESTARTER', but calls RECEIVER with one
            argument, the newly created restarter.")
          ))

(define-documented-interface weak-utilities-interface
  "Sets and hash tables that do not strongly hold their contents; that
   is, if there are no other references to an object contained within a
   weak set or hash table, that object may be destroyed by the garbage
   collector, and it may no longer be an element of the weak set or
   hash table.  These weak sets resemble the built-in populations
   facility in Scheme48, but they are more general.  Weak sets cannot
   contain `#F' (because the value of a weak pointer becomes `#F' if it
   is garbage-collected).  The weak hash tables are only tail-weak;
   they are simple utilities built atop Scheme48's existing hash table
   library."
  (export ((MAKE-WEAK-SET) :weak-set "Weak set constructor.")
          ((WEAK-SET? obj) :boolean "Weak set disjoint type predicate.")
          ((WEAK-SET-EMPTY? weak-set) :boolean
           "Returns true if WEAK-SET is empty and false if it has any
            elements.")
          ((WEAK-SET-ADD! weak-set obj)
           "Adds OBJ to WEAK-SET if it is not already in WEAK-SET.")
          ((WEAK-SET-DELETE! weak-set obj)
           "Deletes a reference to OBJ from WEAK-SET if one exists.")
          ((WEAK-SET-POP! weak-set) :value
           "Removes one element from WEAK-SET and returns it, or `#F'
            if WEAK-SET is empty.  (Yes, the axiom of choice is true of
            these weak sets!)")
          ((WEAK-SET-FOLD weak-set init-seed combiner) :value
           "Fundamental weak set iterator.")
          ((WEAK-SET-FOLD* weak-set init-seed combiner) :value
           "Variant of `WEAK-SET-FOLD' whereby the COMBINER returns two
            values: a boolean, where true signifies that the iteration
            should proceed and false signifies that it should halt; and
            the next seed or, if the boolean is false, the final
            value.")
          ((WEAK-SET-WALK weak-set procedure)
           ;; (put 'weak-set-walk 'scheme-indent-function 1)
           "Applies PROCEDURE to each element of WEAK-SET.")
          ((WEAK-SET-FIND weak-set predicate) :value
           "Returns the first value in WEAK-SET that satisfies
            PREDICATE, or `#F' if no value satisfies it.")
          ((WEAK-SET-CONTAINS? weak-set object) :boolean
           "Returns true if WEAK-SET contains OBJECT and false if not.
            Elements of WEAK-SET are compared to OBJECT with
            `(SCHEME)EQ?'.")
          ((WEAK-SET->LIST weak-set) :list
           "Returns a list of all elements in WEAK-SET.")

          ((MAKE-WEAK-TABLE &optional hash-function) :table
           "Returns a tail-weak hash table with an optionally specified
            hash function that defaults to an `(SCHEME)EQV?'-based
            hash function defined on any values that can usefully be
            used in a `(SCHEME)CASE' clause.")
          ((WEAK-TABLE-REF table key) :value
           "Returns the value associated with KEY in TABLE or `#F' if
            there is no such association.")
          ((WEAK-TABLE-SET! table key value)
           "Associates VALUE with KEY in TABLE.  Subsequent calls to
            `WEAK-TABLE-REF' with the same table and key may not
            reliably produce VALUE, because a garbage collection may
            have occurred and VALUE may have been garbage-collected.")
          ((WEAK-TABLE-WALK procedure table)
           ;; (put 'weak-table-walk 'scheme-indent-function 1)
           "Applies PROCEDURE to two arguments, a key and a value, for
            every association in TABLE.")
          ))