;;; -*- Mode: Scheme; scheme48-package: thread-levels -*-

;;;; Thread Levels

;;; This code is written by Taylor R. Campbell and placed in the Public
;;; Domain.  All warranties are disclaimed.

(define-record-type* level
  (%make-level data
               key
               parent
               continuation
               condition
               pusher-thread
               dynamic-env
               initial-thread
               rebooter
               winder
               unwinder
               )
  ((number (if parent (+ 1 (level-number parent)) 0))
   (child #f)
   (threads (make-weak-table))
   (queue (make-queue))
   (counter (make-counter))
   ))

(define-record-discloser :level
  (lambda (level)
    (list 'level (level-number level) (level-data level))))

(define (make-level data key parent continuation
                    condition pusher-thread dynamic-env
                    initial-thread rebooter winder unwinder)
  (let ((level (%make-level data key parent continuation
                            condition pusher-thread dynamic-env
                            initial-thread rebooter winder unwinder)))
    (spawn-on-level level initial-thread)
    level))

(define     thread-level      thread-data)
(define set-thread-level! set-thread-data!)

(define (level-thread-for-id level uid)
  (weak-table-ref (level-threads level) uid))

(define (walk-level-threads level procedure)
  (weak-table-walk (level-threads level)
                   (lambda (uid thread)
                     uid                ;ignore
                     (procedure thread))))

(define (level-running-thread-count level)
  (counter-value (level-counter level)))

(define (spawn-on-level thread level)
  (set-thread-dynamic-env! thread (level-dynamic-env level))
  (set-thread-scheduler! thread
                         (level-pusher-thread (%top-level level)))
  (set-thread-level! thread level)
  (weak-table-set! (level-threads level)
                   (thread-uid thread)
                   thread)
  (enqueue! (level-queue level) thread)
  (increment-counter! (level-counter level)))

(define (start-top-level data
                         condition pusher-thread make-dynamic-env
                         initial-thread rebooter winder unwinder)
  (let ((key (new-level-key)))
    (%%push-level data key #f
                  condition pusher-thread
                  (with-dynamic-env (make-dynamic-env key)
                    (lambda ()
                      (let-fluid $level-key key
                        get-dynamic-env)))
                  initial-thread rebooter winder unwinder)))

(define (%push-level data parent
                     condition pusher-thread
                     initial-thread rebooter winder unwinder)
  (%%push-level data (level-key parent) parent
                condition pusher-thread
                (thread-dynamic-env pusher-thread)
                initial-thread rebooter winder unwinder))

(define (%%push-level data key parent
                      condition pusher-thread dynamic-env
                      initial-thread rebooter winder unwinder)
  (reify-continuation
   (lambda (continuation)
     (%%%push-level
      (make-level data key parent continuation
                  condition pusher-thread
                  (with-dynamic-env dynamic-env
                    (lambda ()
                      (with-level-restarters parent
                        get-dynamic-env)))
                  initial-thread rebooter winder unwinder)))))

(define (%%%push-level level)
  (let-fluid $level level
    (lambda ()
      (let ((parent (level-parent level)))
        (dynamic-wind
          (lambda ()
            (if parent (set-level-child! parent level))
            ((level-winder level)))
          (lambda ()
            (schedule-level #f))
          (lambda ()
            (unwind-level level (level-unwinder level))
            (if parent (set-level-child! parent #f))))))))

(define (unwind-level level unwinder)
  (let ((unwound? #f))
    (walk-level-threads level
      (lambda (thread)
        (if (thread-continuation thread)
            ;; TERMINATE-THREAD! lets the thread run any unwind thunks,
            ;; so we make sure that they're all on queued up first.
            (let ((queue (level-queue level)))
              (if (not (on-queue? queue thread))
                  (enqueue! queue thread))
              (terminate-thread! thread)))))
    (dynamic-wind
      (lambda ()
        (if unwound?
            (error "trying to throw back into unwound level" level)))
      (lambda ()
        (schedule-level level #t))
      (lambda ()
        (unwinder)
        (set! unwound? #t)))))

;;;; Scheduling

(define (schedule-level level unwinding?)
  (run-threads
   (round-robin-event-handler (level-queue level)
                              level-thread-quantum
                              ;; This dynamic environment argument is
                              ;; used only by the default handler for
                              ;; the SPAWNED event, which we override.
                              #f
                              (level-counter level)
                              (level-event-handler level unwinding?)
                              (level-upcall-handler level)
                              (level-waiter level unwinding?))))

(define level-thread-quantum 200)

(define (level-event-handler level unwinding?)
  (lambda (event arguments)
    (destructure-enum-case event-type (cons event args)
      ((SPAWNED thread)
       (spawn-on-level thread level)
       #t)
      ((RUNNABLE thread)
       (run-level-on-thread thread level)
       #t)
      ((DEADLOCK)
       (if unwinding?
           (warn "deadlock will unwinding level's threads"
                 level))
       (signal 'DEADLOCK)
       #t)
      (else #f))))

(define (run-thread-on-level thread level)
  (let ((other-level (thread-level thread)))
    (if (level? other-level)
        (let loop ((level level))
          (cond ((not level)
                 (warn "dropping thread from exited level"
                       thread
                       other-level))
                ((eq? other-level level)
                 (enqueue! (level-queue level) thread))
                (else
                 (loop (level-parent level)))))
        (error "non-level thread running on a level"
               thread
               level))))

(define (level-waiter level unwinding?)
  (lambda ()
    ;; This returns true if the scheduler has waited and should proceed
    ;; or false if it should terminate.
    (cond ((positive? (level-running-thread-count level))
           (wait)                   ; Call the thread system's waiter.
           #t)
          (unwinding? #f)
          ((level-initial-thread level)
           => (lambda (thread)
                (maybe-reboot-level-initial-thread level thread)
                #t))
          (else #f))))

;;;; Upcalls to the Level Scheduler

;;; Upcalls allow procedures to be called from inferior threads in the
;;; dynamic context of the Swank scheduler thread, where most of the
;;; important information, such as the current level, is stored.

(define (level-upcall-handler level)
  (lambda (origin-thread token arguments)
    (if (and (level-upcall? token)
             (eq? (car arguments) (level-key level)))
        (apply (level-upcall-procedure token)
               origin-thread level (cdr arguments))
        ;; Bloody spelling errors set in stone...
        (propogate-upcall origin-thread token arguments))))

(define-record-type* level-upcall
  (make-level-upcall id procedure)
  ())

(define-record-discloser :level-upcall
  (lambda (token)
    (list 'level-upcall (level-upcall-id token))))

(define-syntax define-level-upcall
  (syntax-rules ()
    ((DEFINE-LEVEL-UPCALL (id parameter ...) thread-var level-var
       body0 body1 ...)
     (DEFINE id
             (LET* ((PROC (LAMBDA (thread-var level-var parameter ...)
                            body0 body1 ...))
                    (TOKEN (MAKE-LEVEL-UPCALL 'id PROC)))
               (LAMBDA (parameter ...)
                 (COND ((FLUID $LEVEL)
                        => (LAMBDA (LEVEL)
                             (PROC (CURRENT-THREAD)
                                   LEVEL
                                   parameter ...)))
                       ((FLUID $LEVEL-KEY)
                        => (LAMBDA (KEY)
                             (UPCALL TOKEN KEY parameter ...)))
                       (ELSE
                        (ERROR "invalid context for level upcall"
                               `(,id ,parameter ...))))))))))

; (put 'define-level-upcall 'scheme-indent-function 3)

(define $level (make-fluid #f))

(define $level-key (make-fluid #f))

(define-level-upcall (current-level) thread level
  thread                                ;ignore
  level)

(define (top-level) (%top-level (current-level)))

(define (%top-level level)
  (cond ((level-parent level) => %top-level)
        (else level)))

(define-level-upcall (current-level-threads) thread level
  thread                                ;ignore
  (cond ((level-parent level) => level-threads)
        (else #f)))

(define (push-level data condition
          initial-thread rebooter
          winder unwinder)
  (let ((thunk
         (push-level/upcall data condition
                            initial-thread rebooter
                            winder unwinder)))
    (thunk)))

; (put 'push-level 'scheme-indent-function 2)

(define-level-upcall (push-level/upcall data condition
                                        initial-thread rebooter
                                        winder unwinder)
    pusher-thread parent-level
  (%push-level data parent-level condition pusher-thread
               initial-thread rebooter
               winder unwinder))

;;; Call only from within the level scheduler.

(define (throw-to-level level thunk)
  (throw-to-continuation (level-continuation level)
                         thunk))

; (put 'throw-to-level 'scheme-indent-function 1)

(define-level-upcall (pop-level) thread level
  thread                                ;ignore
  (cond ((level-parent level) => %resume-level)
        (else (throw-to-level level (lambda () values)))))

(define-level-upcall (resume-level level) thread current
  thread                                ;ignore
  (if (not (eq? level current))
      (%resume-level level)))

(define (%resume-level level)
  (let ((child (level-child level)))
    (cond ((level-pusher-thread (level-child level))
           => (lambda (thread)
                (terminate-thread! thread)
                (if (eq? thread (level-initial-thread level))
                    (maybe-reboot-level-initial-thread level thread))))
          (else
           (warn "resuming level without a pusher thread"
                 level)))
    (throw-to-level child
      (lambda ()
        (lambda ()
          (error "returned to terminated thread"
                 `(after resuming ,level)))))))

(define (maybe-reboot-level-initial-thread level initial-thread)
  ;; INITIAL-THREAD is assumed to have been terminated already.
  (cond ((level-rebooter level)
         => (lambda (rebooter)
              (let ((replacement
                     (make-thread rebooter
                                  (thread-name initial-thread))))
                (spawn-on-level level replacement)
                (set-level-initial-thread! level replacement))))
        (else
         (set-level-initial-thread! level #f))))

(define-level-upcall (reboot-level level initial-thread) thread current
  (if (not (or initial-thread
               (level-rebooter level)))
      (error "level is non-rebootable" level)
      (throw-to-level level
        (lambda ()
          (%%push-level
           (level-data level)
           (level-parent level)
           (level-condition level)
           (level-pusher-thread level)
           (level-dynamic-env level)
           (or initial-thread
               (make-thread (level-rebooter level)
                            (cond ((level-initial-thread level)
                                   => thread-name)
                                  (else #f))))
           (level-rebooter level)
           (level-winder level)
           (level-unwinder level))))))

(define-level-upcall (terminate-levels) thread level
  thread                                ;ignore
  (cond ((level-initial-thread level)
         => (lambda (initial-thread)
              (set-level-initial-thread! level #f)
              (terminate-thread! initial-thread)))))
