;;; -*- mode: scheme; scheme48-package: inspector -*-

;;;;;; Object inspector framework

;;; This code is written by Taylor Campbell and placed in the Public
;;; Domain.  All warranties are disclaimed.

(define-record-type* inspector
  (make-inspector (object))
  ((stack '())
   (history (let ((history (make-xvector)))
              (xvector-set! history 0 object)
              history))
   (history-index 0)))

(define (inspector-push inspector obj)
  (xvector-maybe-extend inspector obj eq?)
  (set-inspector-stack! inspector
                        (cons (inspector-object inspector)
                              (inspector-stack inspector)))
  (set-inspector-object! inspector obj))

(define (inspector-pop inspector)
  (let ((stack (inspector-stack inspector)))
    (cond ((null? stack)
           #f)
          (else
           (set-inspector-stack! inspector (cdr stack))
           (set-inspector-object! inspector (car stack))
           #t))))

(define (inspector-select inspector selector)
  (cond ((procedure? selector)
         (inspector-push inspector
                         (selector (inspector-object inspector)))
         #t)
        ((select-subobject (inspector-object inspector) selector)
         => (lambda (spec)
              (inspector-push inspector (cdr spec))
              #t))
        (else #f)))

(define (inspector-listing inspector position count)
  (inspect-object (inspector-object inspector)
                  position
                  count))

(define-generic object-inspection-title &object-inspection-title
  (obj))
(define-method &object-inspection-title (obj) #f)

(define-generic inspect-object &inspect-object
  (obj position count))
(define-method &inspect-object (obj pos count) #f)

(define-generic select-subobject &select-subobject
  (obj selector))

(define-simple-type :exact-nonnegative-integer (:integer)
  (lambda (x)
    (and (integer? x) (exact? x) (<= 0 x))))

(define-method &select-subobject (obj sel) #f)

;;; These two methods work for simple objects.

(define-method &select-subobject (obj (sel :exact-nonnegative-integer))
  (cond ((inspect-object obj sel 1)
         => (lambda (listing)
              (and (pair? listing)
                   (car listing))))
        (else #f)))

;** Override this method if the listing can be long.

(define-method &select-subobject (obj (sel :symbol))
  (cond ((inspect-object obj sel #f)
         => (lambda (listing)
              (assq sel listing)))
        (else #f)))

(define (maybe-sublist list pos count)
  (let drop ((l list) (i 0))
    (cond ((null? l) #f)
          ((= i pos)
           (if count
               (let take ((l l) (i 0) (tail '()))
                 (if (or (null? l) (= i count))
                     (if (null? tail) #f (reverse tail))
                     (take (cdr l) (+ i 1) (cons (car l) tail))))
               l))
          (else
           (drop (cdr l) (+ i 1))))))



;;; Inspectors & selectors for simple data

(define-method &object-inspection-title ((loc :location))
  "A location (top-level variable slot).")
(define-method &inspect-object ((loc :location) pos count)
  (let ((id (location-id loc)))
    (maybe-sublist `((id . ,id)
                     ,@(let ((name (location-name obj)))
                         (if (and name (not (eq? name id)))
                             `((name . ,name)
                               ,@(cond ((location-package-name loc)
                                        => (lambda (pname)
                                             `((package . ,pname)))
                                        (else '()))))
                             '()))
                     (contents . ,(contents loc)))
                   pos count)))

(define-method &object-inspection-title ((weak :weak-pointer))
  "A weak pointer.")
(define-method &inspect-object ((weak :weak-pointer) pos count)
  (if (zero? pos)
      `((value . ,(weak-pointer-ref weak)))
      #f))

(define-method &object-inspection-title ((cell :cell))
  "A cell.")
(define-method &inspect-object ((cell :cell) pos count)
  (if (zero? pos)
      `((contents . ,(if (cell-unassigned? cell)
                         (unassigned-token)
                         (cell-ref cell))))
      #f))



;;; Indexed object inspection

(let ((define-indexed-inspector
        (lambda (:type string ref length unassigned?)
          (define-method &object-inspection-title ((obj :type))
            string)
          (define-method &inspect-object ((obj :type) pos count)
            (let ((len (length obj)))
              (if (>= pos len)
                  #f
                  (let ((limit (min len (+ pos count))))
                    (let loop ((i pos) (entries '()))
                      (if (= i limit)
                          (reverse entries)
                          (loop (+ i 1)
                                (cons (cons #f (if (unassigned? obj i)
                                                   (unassigned-token)
                                                   (ref obj i)))
                                      entries))))))))
          (define-method &select-subobject
              ((obj :type) (sel :exact-nonnegative-integer))
            (if (< sel (length obj))
                (cons #f (ref obj sel))
                #f))
          (define-method &select-subobject ((obj :type) sel) #f))))
  (define-indexed-inspector :vector "A vector."
    vector-ref vector-length
    vector-unassigned?)
  (define-indexed-inspector :template
    "A template (compiled code component of closures)."
    template-ref template-length
    (lambda (template index) #f)))

(define-record-type* unassigned-token
  (make-unassigned-token)
  ())

(define unassigned-token
        (let ((token (make-unassigned-token)))
          (lambda () token)))

(define-record-discloser type/unassigned-token
  (lambda (token) '(unassigned)))



;;; Pairs & lists

(define-method &object-inspection-title ((nil :null)) "The empty list.")
(define-method &inspect-object ((nil :null) pos count) #f)
(define-method &select-subobject ((nil :null) sel) #f)

(define-method &object-inspection-title ((pair :pair))
  (receive (type len) (careful-length pair)
    (case type
      ((proper) "A proper list.")
      ((dotted) "A dotted list.")
      ((circular) "A circular list."))))

(define-method &inspect-object ((pair :pair) pos count)
  (let ((d (cdr pair)))
    (if (not (pair? d))
        (maybe-sublist `((car . ,(car pair))
                         (cdr . ,(cdr pair)))
                       pos count)
        (receive (type len) (careful-length pair)
          (case type
            ((proper)
             (map (lambda (x) (cons #f x))
                  (maybe-sublist pair pos count)))
            ((dotted)
             (append (map (lambda (x) (cons #f x))
                          (maybe-sublist pair pos count))
                     (list (cons 'dotted-cdr (dotted-cdr pair)))))
            ((circular)
             (map (lambda (x) (cons #f x))
                  (maybe-sublist pair pos count))))))))

(define-method &select-subobject
    ((pair :pair) (sel :exact-nonnegative-integer))
  (if (pair? (cdr pair))
      (receive (type len) (careful-length pair)
        (case type
          ((proper dotted)
           (if (< sel len)
               (cons #f (list-ref pair sel))
               #f))
          ((circular)
           (cons #f (list-ref pair sel)))))
      (next-method)))

(define-method &select-subobject ((pair :pair) (sel :symbol))
  (if (eq? sel 'dotted-cdr)
      (receive (type len) (careful-length pair)
        (if (eq? type 'dotted)
            (cons 'dotted-cdr (dotted-cdr pair))
            #f))
      ;; Silly frill that allows arbitrarily nested CADDADADDDAADADR
      ;; compositions.
      (let* ((string (symbol->string sel))
             (len (string-length string)))
        (and (> len 2)
             (memv (string-ref string 0) '(#\c #\C))
             (memv (string-ref string (- len 1)) '(#\r #\R))
             (let loop ((pair pair) (i 1))
               (let ((j (+ i 1)))
                 (cond ((= j len)
                        (cons sel pair))
                       ((pair? pair)
                        (case (string-ref string i)
                          ((#\a #\A) (loop (car pair) j))
                          ((#\d #\D) (loop (cdr pair) j))
                          (else #f)))
                       (else #f))))))))

(define (careful-length list)
  (let loop ((fast list) (slow list) (len 0))
    (let ((fast-d (cdr fast)))
      (if (pair? fast-d)
          (let ((fast-dd (cdr fast-d))
                (slow-d (cdr slow)))
            (if (pair? fast-dd)
                (if (eq? fast-dd slow-d)
                    (values 'circular len)
                    (loop fast-dd fast-d (+ len 2)))
                (values (if (null? fast-dd) 'proper 'dotted)
                        (+ len 2))))
          (values (if (null? fast-d) 'proper 'dotted)
                  (+ len 1))))))

(define (dotted-cdr list)
  (if (pair? list)
      (dotted-cdr (cdr list))
      list))

;++ closures, continuations, records, ports
